#include <proto.h>

#include "gtfont.h"
#include "gterror.h"
#include "gtconfig.h"

/************************************************************************\
*
* GtFont class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

static GtFont *first=NULL;

/************************************************************************\
*
* Create and init shared font.
*
\************************************************************************/
GtFont *GtMkFont(GtDisplay *d,GtTag *tgs)
{
	GtTag *t,*tt;
	GtFont *obj,*f;
	GtDisplay *fdsp=NULL;
	char *ffname=NULL;
	
	fdsp=d;
	tt=tgs;
	while((t=GtNextTag(&tt)))
	{
		switch(t->tag)
		{
			case GT_Display:
				fdsp=(GtDisplay *)t->d.ptr;
				break;
			case GT_FontList:
				ffname=(char *)t->d.ptr;
				break;
		}
	}
	if(!ffname)
		ffname=GTDEF_FONT;
	for(f=first; f; f=f->next)
	{
		if((fdsp==f->dsp) && (strcmp(ffname,f->fname)==0))
		{
			f->cnt++;
			return(f);
		}
	}
	if(!(obj=new GtFont()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtFont object!\n");
		return(NULL);
	}
	obj->next=first;
	first=obj;
	obj->cnt=0;
	if(!obj->Init(d,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init shared font.
*
\************************************************************************/
GtFont *GtMkFont(GtDisplay *d,ulong tag,...)
{
	va_list va;
	GtFont *obj;
	int p;
	
	va_start(va,tag);
	p=GtBeginVaTags();
	obj=GtMkFont(d,GtCopyVaTags(tag,&va));
	GtEndVaTags(p);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Delete shared font.
*
\************************************************************************/
void GtFont::DeInit(void)
{
	GtFont *f;
	
	if(cnt>0)
	{
		cnt--;
		return;
	}
	if(cnt==0)
	{
		f=first;
		if(f==this)
			first=next;
		else
		{
			while(f && (f->next!=this))
				f=f->next;
			if(f)
				f->next=next;
		}
	}
	GtBase::DeInit();
}

/************************************************************************\
*
* Create empty font.
*
\************************************************************************/
GtFont::GtFont(void)
{
	dsp=NULL;
	next=NULL;
	fs=NULL;
	fnt=0;
	fname=NULL;
	cnt=0;
	nofreex=FALSE;
}

/************************************************************************\
*
* Destroy font.
*
\************************************************************************/
void GtFont::Destroy(void)
{
	if(fs && !nofreex)
		XFreeFont(dsp->XDsp(),fs);
	GtBase::Destroy();
}

/************************************************************************\
*
* Setup font.
*
\************************************************************************/
sbool GtFont::Init(GtDisplay *d,GtTag *tgs)
{
	GtTag *t;
	char *fl,*nf;
	
	dsp=d;
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_FontList:
				fname=(char *)t->d.ptr;
				break;
			case GT_XFont:
				fnt=(Font)t->d.ptr;
				nofreex=TRUE;
				break;
			case GT_DontFreeX:
				if(t->d.val)
					nofreex=TRUE;
				else
					nofreex=FALSE;
				break;
		}
	}
	if(!dsp)
	{
		GtError(this,GTEC_NODISPLAY,"No GtDisplay!\n");
		return(FALSE);
	}
	if(!fnt)
	{
		if(!fname)
			fname=GTDEF_FONT;
		if(!(fl=(char *)malloc(strlen(fname)+1)))
		{
			GtError(this,GTEC_NOMEMORY,"No memory for font list!\n");
			return(FALSE);
		}
		strcpy(fl,fname);
		for(nf=strtok(fl,",:"); nf; nf=strtok(NULL,",:"))
		{
			if((fs=XLoadQueryFont(dsp->XDsp(),nf)))
				break;
		}
		if(!fs)
		{
			GtError(this,GTEC_NOXRESOURCE,"Unable to load font from list: %s\n",fname);
			return(FALSE);
		}
		fnt=fs->fid;
		free(fl);
	}
	else
	{
		if(!(fs=XQueryFont(dsp->XDsp(),fnt)))
		{
			GtError(this,GTEC_NOXRESOURCE,"Unable to find information about font!\n");
			return(FALSE);
		}
	}
	return(TRUE);
}

/************************************************************************\
*
* Find display tag and then setup normal.
*
\************************************************************************/
sbool GtFont::Init(GtTag *tgs)
{
	GtDisplay *d;
	
	d=(GtDisplay *)GtGetTagPtr(tgs,GT_Display,NULL);
	return(this->Init(d,tgs));
}

/************************************************************************\
*
* Set font attributes.
*
\************************************************************************/
sbool GtFont::Set(GtTag *tgs)
{
	tgs=tgs;
	return(TRUE);
}

/************************************************************************\
*
* Get font attributes.
*
\************************************************************************/
void GtFont::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;

	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Display:
				*ptr=(long)dsp;
				break;
			case GT_FontList:
				*ptr=(long)fname;
				break;
			case GT_XFont:
				*ptr=(long)fnt;
				break;
			case GT_DontFreeX:
				*ptr=nofreex;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtFont::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtFont")==0)
		return(TRUE);
	else
		return(GtBase::IsSubClassOf(cn));
}

/************************************************************************\
*
* Get string extents.
*
\************************************************************************/
void GtFont::StrExtents(char *str,int len,XCharStruct *xcs)
{
	int d1,d2,d3;
	
	XTextExtents(fs,str,len,&d1,&d2,&d3,xcs);
}

/************************************************************************\
*
* Get minimum font extents.
*
\************************************************************************/
void GtFont::MinExtents(XCharStruct *xcs)
{
	memcpy(xcs,&fs->min_bounds,sizeof(XCharStruct));
}

/************************************************************************\
*
* Get maximum font extents.
*
\************************************************************************/
void GtFont::MaxExtents(XCharStruct *xcs)
{
	memcpy(xcs,&fs->max_bounds,sizeof(XCharStruct));
}

/************************************************************************\
*
* Calculate how many characters from string fits in given width.
*
\************************************************************************/
int GtFont::StrFit(char *str,int len,int xsize,int *fxs)
{
	int n,w=0;
	
	*fxs=0;
	for(n=0; n<len; n++)
	{
		w+=this->CharWidth(str[n]);
		if(w<xsize)
			*fxs=w;
		else
			break;
	}
	return(n);
}
