#include <proto.h>

#include "gtpanner.h"
#include "gtknob.h"
#include "gtpulseknob.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtresname.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtPanner class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

/************************************************************************\
*
* Create and init panner.
*
\************************************************************************/
GtPanner *GtMkPanner(char *n,GtManager *p,GtTag *tgs)
{
	GtPanner *obj;
	
	if(!(obj=new GtPanner()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtPanner object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init panner.
*
\************************************************************************/
GtPanner *GtMkPanner(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtPanner *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkPanner(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Setup empty panner.
*
\************************************************************************/
GtPanner::GtPanner(void)
{
	maxchild=1;
	knob=NULL;
	xstart=ystart=0;
	xrange=yrange=100;
	xpage=ypage=1;
	xval=yval=0;
	xknpos=yknpos=0;
	xknsz=yknsz=0;
	kmxs=1;
	kmys=1;
	sflags=0;
	ibw=0;
	bdrw=1;
	ibdrw=0;
	ftype=GTF_DBL_OUT;
}

/************************************************************************\
*
* Setup panner.
*
\************************************************************************/
sbool GtPanner::Init(char *n,GtManager *p,GtTag *tgs)
{
	GtTag *t;
	char *kname="knob";
	long sx=0,sy=0,ex=100,ey=100,del=0;
	
	if(!GtManager::Init(n,p,tgs))
		return(FALSE);
	app->GetRes(this,GTR_FRAME,GTRT_FRAME,&ftype);
	app->GetRes(this,GTR_IBDR,GTRT_BYTE,&ibdrw);
	app->GetRes(this,GTR_BDR,GTRT_BYTE,&bdrw);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Start:
				sx=(long)t->d.val;
				sy=(long)t->d.val;
				break;
			case GT_End:
				ex=(long)t->d.val;
				ey=(long)t->d.val;
				break;
			case GT_Value:
				xval=t->d.val;
				yval=t->d.val;
				break;
			case GT_Page:
				xpage=t->d.val;
				ypage=t->d.val;
				break;
			case GT_StartX:
				sx=(long)t->d.val;
				break;
			case GT_EndX:
				ex=(long)t->d.val;
				break;
			case GT_ValueX:
				xval=t->d.val;
				break;
			case GT_PageX:
				xpage=t->d.val;
				break;
			case GT_StartY:
				sy=(long)t->d.val;
				break;
			case GT_EndY:
				ey=(long)t->d.val;
				break;
			case GT_ValueY:
				yval=t->d.val;
				break;
			case GT_PageY:
				ypage=t->d.val;
				break;
			case GT_ReportDrag:
				if(t->d.val)
					sflags|=GTPF_REPDRAG;
				else
					sflags&=~GTPF_REPDRAG;
				break;
			case GT_ReduceInterval:
				if(t->d.val)
					sflags|=GTPF_REDUCE;
				else
					sflags&=~GTPF_REDUCE;
				break;
			case GT_KnobName:
				kname=(char *)t->d.ptr;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				break;
			case GT_Border:
				bdrw=t->d.val;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_DBL_OUT;
				break;
			case GT_Delay:
				del=t->d.val;
				break;
			case GT_Pulse:
				if(t->d.val)
					sflags|=GTPF_PULSE;
				else
					sflags&=~GTPF_PULSE;
				break;
		}
	}
	switch(ftype)
	{
		case GTF_NONE:
		case GTF_SIMPLE:
		case GTF_IN:
		case GTF_OUT:
			ibw=ibdrw+bdrw;
			break;
		case GTF_DBL_IN:
		case GTF_DBL_OUT:
			ibw=ibdrw+2*bdrw;
			break;
	}
	if(sflags & GTPF_PULSE)
	{
		if((sx>0) && (ex>0))
			sx=0;
		if((sx<0) && (ex<0))
			ex=0;
		if((sy>0) && (ey>0))
			sy=0;
		if((sy<0) && (ey<0))
			ey=0;
	}
	xstart=GTMIN(sx,ex);
	ex=ex-sx;
	if(ex<0)
		xrange=-ex;
	else
		xrange=ex;
	if(xpage>xrange)
		xpage=xrange;
	if(sflags & GTPF_REDUCE)
		xrrange=xrange-xpage;
	else
		xrrange=xrange;
	if(sflags & GTPF_PULSE)
		xval=0;
	else
		xval=GTMAX(xstart,GTMIN(xstart+(long)xrrange,xval));
	ystart=GTMIN(sy,ey);
	ey=ey-sy;
	if(ey<0)
		yrange=-ey;
	else
		yrange=ey;
	if(ypage>yrange)
		ypage=yrange;
	if(sflags & GTPF_REDUCE)
		yrrange=yrange-ypage;
	else
		yrrange=yrange;
	if(sflags & GTPF_PULSE)
		yval=0;
	else
		yval=GTMAX(ystart,GTMIN(ystart+(long)yrrange,yval));
	if(kname)
	{
		if(sflags & GTPF_PULSE)
		{		
			if(!(knob=GtMkPulseKnob(kname,this,
					GT_Disabled,(flags & GTWF_DISABLED ? TRUE : FALSE),
					GT_Delay,del,
					GT_DONE)))
				return(FALSE);
		}
		else
		{
			if(!(knob=GtMkKnob(kname,this,
					GT_Disabled,(flags & GTWF_DISABLED ? TRUE : FALSE),
					GT_DONE)))
				return(FALSE);
		}
	}
	return(TRUE);
}

/************************************************************************\
*
* Set panner attributes.
*
\************************************************************************/
sbool GtPanner::Set(GtTag *tgs)
{
	GtTag *t;
	long sx,sy,ex,ey;
	sbool och=FALSE,ch=FALSE,rch=FALSE,vch=FALSE;

	if(!GtManager::Set(tgs))
		return(FALSE);
	sx=xstart;
	ex=xstart+xrange;
	sy=ystart;
	ey=ystart+yrange;
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Start:
				sx=(long)t->d.val;
				sy=(long)t->d.val;
				rch=TRUE;
				break;
			case GT_End:
				ex=(long)t->d.val;
				ey=(long)t->d.val;
				rch=TRUE;
				break;
			case GT_Value:
				xval=t->d.val;
				yval=t->d.val;
				ch=TRUE;
				break;
			case GT_Page:
				xpage=t->d.val;
				ypage=t->d.val;
				rch=TRUE;
				break;
			case GT_StartX:
				sx=t->d.val;
				rch=TRUE;
				break;
			case GT_EndX:
				ex=t->d.val;
				rch=TRUE;
				break;
			case GT_ValueX:
				xval=t->d.val;
				ch=TRUE;
				break;
			case GT_PageX:
				xpage=t->d.val;
				rch=TRUE;
				break;
			case GT_StartY:
				sy=t->d.val;
				rch=TRUE;
				break;
			case GT_EndY:
				ey=t->d.val;
				rch=TRUE;
				break;
			case GT_ValueY:
				yval=t->d.val;
				ch=TRUE;
				break;
			case GT_PageY:
				ypage=t->d.val;
				rch=TRUE;
				break;
			case GT_ReportDrag:
				if(t->d.val)
					sflags|=GTPF_REPDRAG;
				else
					sflags&=~GTPF_REPDRAG;
				break;
			case GT_ReduceInterval:
				if(t->d.val)
					sflags|=GTPF_REDUCE;
				else
					sflags&=~GTPF_REDUCE;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_Border:
				bdrw=t->d.val;
				vch=TRUE;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_DBL_OUT;
				vch=TRUE;
				break;
			case GT_Delay:
				if(knob)
					knob->SetV(GT_Delay,t->d.val,GT_DONE);
				break;
		}
	}
	if(vch)
	{
		switch(ftype)
		{
			case GTF_NONE:
			case GTF_SIMPLE:
			case GTF_IN:
			case GTF_OUT:
				ibw=ibdrw+bdrw;
				break;
			case GTF_DBL_IN:
			case GTF_DBL_OUT:
				ibw=ibdrw+2*bdrw;
				break;
		}
	}
	if(rch || ch)
	{
		if(rch)
		{
			if(sflags & GTPF_PULSE)
			{
				if((sx>0) && (ex>0))
					sx=0;
				if((sx<0) && (ex<0))
					ex=0;
				if((sy>0) && (ey>0))
					sy=0;
				if((sy<0) && (ey<0))
					ey=0;
			}
			xstart=GTMIN(sx,ex);
			ex=ex-sx;
			if(ex<0)
				xrange=-ex;
			else
				xrange=ex;
			if(xpage>xrange)
				xpage=xrange;
			if(sflags & GTPF_REDUCE)
				xrrange=xrange-xpage;
			else
				xrrange=xrange;
			ystart=GTMIN(sy,ey);
			ey=ey-sy;
			if(ey<0)
				yrange=-ey;
			else
				yrange=ey;
			if(ypage>yrange)
				ypage=yrange;
			if(sflags & GTPF_REDUCE)
				yrrange=yrange-ypage;
			else
				yrrange=yrange;
		}
		if(sflags & GTPF_PULSE)
			xval=0;
		else
			xval=GTMAX(xstart,GTMIN(xstart+(long)xrrange,xval));
		if(sflags & GTPF_PULSE)
			yval=0;
		else
			yval=GTMAX(ystart,GTMIN(ystart+(long)yrrange,yval));
	}
	if(ch || och || vch)
	{
		this->Relayout();
	}
	return(TRUE);
}

/************************************************************************\
*
* Get panner attributes.
*
\************************************************************************/
void GtPanner::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	GtManager::Get(tgs);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_StartX:
				*ptr=xstart;
				break;
			case GT_EndX:
				*ptr=xstart+xrange;
				break;
			case GT_ValueX:
				*ptr=xval;
				break;
			case GT_PageX:
				*ptr=xpage;
				break;
			case GT_StartY:
				*ptr=ystart;
				break;
			case GT_EndY:
				*ptr=ystart+yrange;
				break;
			case GT_ValueY:
				*ptr=yval;
				break;
			case GT_PageY:
				*ptr=ypage;
				break;
			case GT_ReportDrag:
				if(sflags & GTPF_REPDRAG)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_ReduceInterval:
				if(sflags & GTPF_REDUCE)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_InnerBorder:
				*ptr=ibdrw;
				break;
			case GT_Border:
				*ptr=bdrw;
				break;
			case GT_FrameType:
				*ptr=ftype;
				break;
			case GT_Delay:
				if(knob)
					knob->GetV(GT_Delay,ptr,GT_DONE);
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtPanner::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtPanner")==0)
		return(TRUE);
	else
		return(GtManager::IsSubClassOf(cn));
}

/************************************************************************\
*
* Allow only subclasses of GtKnob to be added.
*
\************************************************************************/
sbool GtPanner::AddChild(GtWidget *o,ulong pos)
{
	if(o && o->Manage())
	{
		if(o->IsSubClassOf("GtKnob"))
		{
			knob=(GtKnob *)o;
			if(knob->IsSubClassOf("GtPulseKnob"))
				sflags|=GTPF_PULSE;
			return(GtManager::AddChild(o,pos));
		}
		else
		{
			GtError(this,GTEC_NOADD,"Unable to add non-knob child: %s\n",o->GetName());
			return(FALSE);
		}
	}
	else
		return(GtManager::AddChild(o,pos));
}

/************************************************************************\
*
* Child asked to be moved and/or resized.
*
\************************************************************************/
void GtPanner::MoveSizeReq(GtWidget *obj,word xp,word yp,word xs,word ys)
{
	GtPannerCD cd;
	word oxs,oys;
	long vx,vy;

	if(flags & GTWF_UNBUILT)
	{
		obj->SetPosSize(xp,yp,xs,ys);
		return;
	}
	oxs=xs;
	oys=ys;
	vx=xval;
	vy=yval;
	if(xp<ibw)
		xp=ibw;
	if(xp+xknsz>xsize-ibw)
		xp=xsize-xknsz-ibw;
	xknpos=xp;
	xs=xknsz;
	if(xsize-xknsz-2*ibw)
		xval=xstart+xrrange*(xknpos-ibw)/(xsize-xknsz-2*ibw);
	if(xknpos==ibw)
		xval=xstart;
	if(xknpos+xknsz==xsize-ibw+1)
		xval=xstart+xrrange;
	xval=GTMAX(xstart,GTMIN(xstart+(long)xrrange,xval));
	if(yp<ibw)
		yp=ibw;
	if(yp+yknsz>ysize-ibw)
		yp=ysize-yknsz-ibw;
	yknpos=yp;
	ys=yknsz;
	if(ysize-yknsz-2*ibw)
		yval=ystart+yrrange*(yknpos-ibw)/(ysize-yknsz-2*ibw);
	if(yknpos==ibw)
		yval=ystart;
	if(yknpos+yknsz==ysize-ibw+1)
		yval=ystart+yrrange;
	yval=GTMAX(ystart,GTMIN(ystart+(long)yrrange,yval));
	if((oxs>kmxs) || (oys>kmys))
		this->Relayout();
	else
		obj->SetPosSize(xp,yp,xs,ys);
	if((sflags & GTPF_REPDRAG) && ((vx!=xval) || (vy!=yval)))
	{
		cd.xvalue=xval;
		cd.yvalue=yval;
		cd.drag=TRUE;
		chg.CallCB(this,&cd);
	}
}

/************************************************************************\
*
* Relayout panner knob.
*
\************************************************************************/
void GtPanner::Relayout(void)
{
	int xs=1,ys=1;
	
	if(flags & GTWF_UNBUILT)
		return;
	if(knob)
		knob->GetMinSize(&xs,&ys);
	kmxs=GTMAX(xs,kmxs);
	kmys=GTMAX(ys,kmys);
	this->SetMinMaxSize(GTMAX(2*ibw+kmxs,2*ibw+1),GTMAX(2*ibw+kmys,2*ibw+1),
											GTS_MAX,GTS_MAX);
}

/************************************************************************\
*
* Set panner event mask.
*
\************************************************************************/
void GtPanner::SetEMask(ulong em)
{
	GtManager::SetEMask(em|ExposureMask|ButtonReleaseMask|ButtonPressMask);
}

/************************************************************************\
*
* Add callback.
*
\************************************************************************/
void GtPanner::AddCB(ulong tag,GtCallB cb,GtUDPtr ud)
{
	if(tag==GT_ValueChange)
		chg.AddCB(cb,ud);
	else
		GtManager::AddCB(tag,cb,ud);
}

/************************************************************************\
*
* Remove callback.
*
\************************************************************************/
void GtPanner::RemCB(ulong tag,GtCallB cb)
{
	if(tag==GT_ValueChange)
		chg.RemCB(cb);
	else
		GtManager::RemCB(tag,cb);
}

/************************************************************************\
*
* Notify from knob release.
*
\************************************************************************/
void GtPanner::Notify(ulong tag)
{
	GtPannerCD cd;
	
	if(tag==GT_Activate)
	{
		cd.xvalue=xval;
		cd.yvalue=yval;
		cd.drag=FALSE;
		chg.CallCB(this,&cd);
	}
	else
		GtManager::Notify(tag);
}

/************************************************************************\
*
* Layout panner and knob.
*
\************************************************************************/
void GtPanner::Layout(void)
{
	int xp,yp,xs,ys;

	if(flags & GTWF_UNBUILT)
		return;
	if(knob)
	{
		knob->GetMinSize(&xs,&ys);
		kmxs=GTMAX(xs,kmxs);
		kmys=GTMAX(ys,kmys);
		if(xrange)
			xknsz=kmxs+(xsize-2*ibw-kmxs)*xpage/xrange;
		else
			xknsz=kmxs+(xsize-2*ibw-kmxs);
		xknsz=GTMAX(xknsz,xs);
		if(xrrange)
			xknpos=ibw+(xsize-2*ibw-xknsz)*(xval-xstart)/xrrange;
		else
			xknpos=ibw;
		xp=xknpos;
		xs=xknsz;
		if(yrange)
			yknsz=kmys+(ysize-2*ibw-kmys)*ypage/yrange;
		else
			yknsz=kmys+(ysize-2*ibw-kmys);
		yknsz=GTMAX(yknsz,ys);
		if(yrrange)
			yknpos=ibw+(ysize-2*ibw-yknsz)*(yval-ystart)/yrrange;
		else
			yknpos=ibw;
		yp=yknpos;
		ys=yknsz;
		knob->SetPosSize(xp,yp,xs,ys);
	}
}

/************************************************************************\
*
* Draw panner.
*
\************************************************************************/
sbool GtPanner::Draw(XEvent *xe)
{
	int n;

	if(xe && ((xe->type!=Expose) || (xe->xexpose.count!=0)))
		return(TRUE);
	switch(ftype)
	{
		case GTF_SIMPLE:
			for(n=0; n<bdrw; n++)
				gtfore->DrawRectangle(gtwin,n,n,xsize-2*n-1,ysize-2*n-1,FALSE);
			break;
		case GTF_DBL_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtlht,gtdrk,gtwin,bdrw+n,bdrw+n,xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_DBL_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtdrk,gtlht,gtwin,bdrw+n,bdrw+n,xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,n,n,xsize-2*n,ysize-2*n);
			}
			break;
		case GTF_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,n,n,xsize-2*n,ysize-2*n);
			}
			break;			
	}
	return(TRUE);
}

/************************************************************************\
*
* Mouse panner events.
*
\************************************************************************/
sbool GtPanner::Mouse(XEvent *xe)
{
	GtPannerCD cd;
	long xp,yp,vx,vy;

	if(flags & GTWF_DISABLED)
		return(FALSE);
	if((xe->type==ButtonRelease) && (xe->xbutton.button==Button1))
	{
		if(sflags & GTPF_PULSE)
			return(TRUE);
		xp=xe->xbutton.x;
		yp=xe->xbutton.y;
		if((xp>ibw) && (xp<xsize-ibw) && (yp>ibw) && (yp<ysize-ibw))
		{
			vx=xval;
			vy=yval;
			if(xp<xknpos)
				xval-=xpage;
			if(xp>xknpos+xknsz)
				xval+=xpage;
			if(yp<yknpos)
				yval-=ypage;
			if(yp>yknpos+yknsz)
				yval+=ypage;
			xval=GTMAX(xstart,GTMIN(xstart+(long)xrrange,xval));
			yval=GTMAX(ystart,GTMIN(ystart+(long)yrrange,yval));
			this->Layout();
			if((vx!=xval) || (vy!=yval))
			{
				cd.xvalue=xval;
				cd.yvalue=yval;
				cd.drag=FALSE;
				chg.CallCB(this,&cd);
			}
		}
		return(TRUE);
	}
	return(FALSE);
}
